#include "Port.h"
#include <Preferences.h> //lib pour écrire dans l'eeprom
#include <arpschuino32_core.h>
#include <arpschuino.h>
#include "custom_port.h"


Port portA('A');  //creation des objets port
Port portB('B');

Preferences port_preferences;

Port::Port(char name):m_port_name(name)
{

}

///////////////////////////////////////////////////////
////////////////OTHER METHODS/////////////////////////
//////////////////////////////////////////////////////

void Port::debug()
{

    Serial.print("debug port ");
    Serial.println(m_port_name);
}
///////////////////////////////////////////////////////
////////////////SET METHODS////////////////////////////
//////////////////////////////////////////////////////

void Port::setNbDmxChannels(uint16_t nbChannels)//should auto adjust, except for pixel Led
    {
        m_nbDmxChannels=nbChannels;
    }
/////////////////////////////////////////////////////////
////////////////RECORD METHODS//////////////////////////
///////////////////////////////////////////////////////
void Port::record(String name,uint8_t val )const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    port_preferences.putUChar(key, val);
}

void Port::record(String name,uint16_t val )const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    port_preferences.putUShort(key, val);
}

void Port::record(String name,uint32_t val )const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    port_preferences.putULong(key, val);
}

void Port::record(String name,bool val )const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    port_preferences.putBool(key, val);
}

void Port::record(String name,float val )const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    port_preferences.putFloat(key, val);
}
/////////////////////////////////////////////////////////////////////////////
//////////////////////READ METHODS (read preferences)/////////////////////
////////////////////////////////////////////////////////////////////////////

uint8_t Port::load(String name, uint8_t val)const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    return port_preferences.getUChar(key, val);
}

uint16_t Port::load(String name,uint16_t val)const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    return port_preferences.getUShort(key, val);
}

uint32_t Port::load(String name,uint32_t val)const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    return port_preferences.getULong(key, val);
}

bool Port::load(String name,bool val)const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    return port_preferences.getBool(key, val);
}

float Port::load(String name,float val)const 
{
    String str(m_port_name); 
    str += name; 
    const char * key = str.c_str();
    return port_preferences.getFloat(key, val);
}
///////////////////////////////////////////////////////
////////////////GET METHODS////////////////////////////
//////////////////////////////////////////////////////

uint16_t Port::getNbDmxChannels()const
{
    return m_nbDmxChannels;
}
///////////////////////////////////////////////////////
//////////////////SETUP METHODS////////////////////////
///////////////////////////////////////////////////////
void Port::init()//renommé
{
    port_preferences.begin("port_datas", false);
    Serial.print("Port ");
    uint8_t defaultVal=0;
    m_mode=load("mode",defaultVal);
    Serial.print(m_port_name);

    switch(m_mode)
    {
        case 0: {  //pwm
            PWM_setup();
            break;
        } 
        case 1:  { 
            on_off_setup();
            break;
        }
        case 2:  { 
            servo_setup();
           
            break;
        }
        case 3:  { 
            stepper_setup();
            break;
        }
        case 4:  { 
            CCmotor_setup();
            break;
        }        
        case 254:  {
            custom_setup();
            break;                           
        }
        default: {
            Serial.println(": inactive...");
        }
    }       
}

void Port::PWM_setup()
{
    int32_t  freq = 40000000;
    uint8_t defaultVal=14;
    m_pwmRes= load("pwmRes",defaultVal);     
    for(uint8_t i=1;i<m_pwmRes;i++){
        freq=freq/2;
    }
    //freq = 1600;//debug ///////////////////////////////////////////////////////////////
    uint8_t startPin= (m_port_name=='A') ? 0 : 8;
    m_dmxRes= load("dmxRes",false);   
    if(m_dmxRes){
        m_nbDmxChannels=16;
    }
    else{
        m_nbDmxChannels=8;
    }
    for(uint8_t i=startPin;i<startPin+8;i++)
    {  
        //pinMode(Arp[i], OUTPUT);  
        ledcSetup(i, freq, m_pwmRes);// configure LED PWM functionalitites
        ledcAttachPin(Arp[i], i);// attach the channel to the GPIO to be controlled
    } 
    Serial.print(": PWM ");
    Serial.print(m_pwmRes);
    Serial.print(" bits; freq ");
    Serial.print(freq);
    Serial.print(" hz");
    if(m_dmxRes)
        Serial.print(", 16 bits DMX input");  
    Serial.println();     
} 

void Port::on_off_setup()
{   
    m_inverted= load("inverted",false);//defaultVal);
    uint8_t defaultVal=128; 
    m_trigger= load("trigger",defaultVal);    
    uint8_t startPin= (m_port_name=='A') ? 0 : 8;
    m_nbDmxChannels=8;
    for(int i=startPin;i<startPin+8;i++)
    {
        pinMode(Arp[i], OUTPUT);
        if(m_inverted)digitalWrite(Arp[i], HIGH);
    } 
    Serial.print(": on/off, trigger at ");
    Serial.print(m_trigger); 
    Serial.print(" (");
    Serial.print((m_trigger*100)/255);
    Serial.print("%)");
    if(m_inverted)
        Serial.print(", inverted");  
    Serial.println(); 
}

void Port::servo_setup()
{
    m_inverted = load("inverted",false);
    uint8_t defaultAngle=160;
    m_servoAngle = load("angle",defaultAngle);
    m_servoDetach = load("detach",false);
    m_nbDmxChannels = 8;
    //uint8_t startPin= 0 ;
    uint8_t startPin= (m_port_name=='A') ? 0 : 8;
    m_servo = new Servo[8];
    if(m_servoDetach==false)
    {
        for(int i=0;i<8;i++)
        {
            uint8_t output=i+startPin;
            m_servo[i].attach(Arp[output],output);//(pin,channel)
        }
    }
    Serial.print(": Servo");
    Serial.print(", ");
    Serial.print(m_servoAngle);
    Serial.print("°");
    if(m_inverted)
        Serial.print(", inverted");
    if(m_servoDetach)
        Serial.print(", mode detach");
    Serial.println();
}
 
void Port:: stepper_setup()
{ 
    Serial.println(" : Stepper");
    uint16_t defaultSPR = 800;
    uint16_t defaultMin = 4;
    uint16_t defaultMax = 200;
    float defaultRevol = 12;
    bool defaultAction = 0; //perform by default

    m_stepper = new ArpStepper[4];
    m_nbDmxChannels = 8;
    uint8_t startPin= (m_port_name=='A') ? 0 : 8;
    for(int i=0;i<4;i++)
    {
        String num(i);
        String str = num+"stpActive";
        const char * word = str.c_str(); 
        m_stp_active[i]= load(word,true);
        Serial.print("stepper ");  Serial.print(i); 
        if(m_stp_active[i])
        {
            str = num+"stpPerRevol"; 
            word = str.c_str(); 
            uint16_t stepPerRevolution = load(word,defaultSPR);
            //uint8_t output=i+startPin;
            m_stepper[i].attach(Arp[i*2+1+startPin],Arp[i*2+startPin],stepPerRevolution);

            Serial.print(" steps per revolution : "); Serial.print(stepPerRevolution); Serial.print("; ");  
            ////////////////////////////////////////////////////////////////////
            str = num+"_rpm_min";
            word = str.c_str();
            uint16_t rpmMin = load(word,defaultMin);
            Serial.print(" rpmmin : "); Serial.print(rpmMin); Serial.print("; "); 

            str = num+"_rpm_max";
            word = str.c_str();
            uint16_t rpmMax = load(word,defaultMax);
            Serial.print(" max : "); Serial.print(rpmMax); Serial.print("; ");   

            str = num+"_revol";
            word = str.c_str();
            float revol = load(word,defaultRevol);
            Serial.print(" revolutions : "); Serial.print(revol); Serial.print("; ");                     


            str = num+"stpAction";
            word = str.c_str();
            m_action[i] = load(word,defaultAction);
            Serial.print(" action : ");
            uint16_t dmxRes = _8bits;

            if(m_action[i]==0)
            {
                Serial.print("perform"); Serial.print("; ");
                Serial.print(" DMX resolution : ");
                str = num+"stp_dmxRes";
                word = str.c_str();
                //Serial.print("((word : ");Serial.print(word);Serial.print(" ))");//////////debug////////
                m_stp_dmxRes[i]=load(word,false);
                //Serial.print("((16bits : ");Serial.print(m_stp_dmxRes[i]);Serial.print(" ))");//////////debug////////
                if(m_stp_dmxRes[i]==true){
                    dmxRes = _16bits;
                    Serial.print("16bits"); 
                    m_nbDmxChannels++;
                } 
                else{
                    Serial.print("8bits");
                }
                Serial.print("; ");
            }
            else 
            {
                Serial.print("continuous; ") ;
            }                

            m_stepper[i].init(rpmMin, rpmMax, revol, dmxRes); 
            //Serial.println();Serial.print(" rpmMin (2): ");Serial.println(rpmMin);

            str = num+"stpInverted";
            word = str.c_str();
            bool inverted = load(word,false);
            if(inverted)
            {
                m_stepper[i].invert_rotation(true);
                Serial.println(" inverted; "); 
            }
            Serial.println(); 
        }
        else Serial.println(" inactive");
    }
    Serial.println();    
}

void Port::CCmotor_setup()
{  
    Serial.println(" : CC motor");
    uint8_t default_CC_threshold = 1;

   for(int i=0; i<4; i++)
    {
        Serial.print("motor ");  Serial.print(i); 
        String num(i);
        String str = num+"thresold";
        const char * word = str.c_str(); 

        m_cc_threshold[i]=load(word,default_CC_threshold);
        Serial.print(" thresold : "); Serial.print(m_cc_threshold[i]);
        uint8_t percent = m_cc_threshold[i]*100/255;
        Serial.print(" ("); Serial.print(percent);Serial.print("%)");

        str = num+"CCInverted";
        word = str.c_str(); 
        m_CC_inverted[i]=load(word,false);
        Serial.print(" CCInverted : "); Serial.println(m_CC_inverted[i]);
    }
    m_nbDmxChannels = 8;    
    uint8_t startPin= (m_port_name=='A') ? 0 : 8;  

    for(int i=startPin;i<startPin+8;i++)
    {
        ledcSetup(i,25000,8);//channel, freq, Resolution,25000Hz should be unaudible
        ledcAttachPin(Arp[i], i);// attach the channel to the GPIO to be controlled
    }      
}

void Port::custom_setup()
{
    if(m_port_name=='A')
    {
        setupA();
    }
    else if(m_port_name=='B')
    {
        setupB();
    }
}

///////////////////////////////////////////////////////
//////////////////ACTION METHODS////////////////////////
///////////////////////////////////////////////////////
void Port:: action(const std::vector<uint8_t> & DMXslice)
{
    uint8_t startPin= (m_port_name=='A') ? 0 : 8;
    switch(m_mode)
    {
        case 0: {  
            PWM_action(DMXslice,startPin);
            break;
        } 
        case 1:  { 
            on_off_action(DMXslice,startPin);
            break;
        }
        case 2:  {
            servo_action(DMXslice,startPin);           
            break;               
        }
        case 3 : {
            stepper_action(DMXslice,startPin);
            break;
        }
        case 4 : {
            CCmotor_action(DMXslice,startPin);
            break;
        }        
        case 254 : {
            custom_action(DMXslice,startPin);
        }
    }
} 

void Port:: PWM_action(const std::vector<uint8_t> & DMXslice,uint8_t startPin)
{
    //uint16_t level;
    if(m_dmxRes){//if 16 bits DMX
        if(m_pwmCurve==0){//if led curve
            for (int i=0; i < 8; ++i) 
            {
                m_target_level[i] = DMXslice[i*2] << 8 | DMXslice[i*2+1];//level= DMXslice[i*2] << 8 | DMXslice[i*2+1];
                uint32_t level32 =((m_target_level[i]+1)*(m_target_level[i]+1))-1;//uint32_t level32 =((level+1)*(level+1))-1;
                m_target_level[i] = level32 >>(32-m_pwmRes);//level = level32 >>(32-m_pwmRes);
                //ledcWrite(startPin+i, level);
            }
        }
        else{//if straight curve
            for (int i=0; i < 8; ++i) 
            {
                m_target_level[i] = DMXslice[i*2] << 8 | DMXslice[i*2+1];//level= DMXslice[i*2] << 8 | DMXslice[i*2+1];
                //ledcWrite(startPin+i, level);
            }
        }            
    }
    else{//if 8 bits DMX
        if(m_pwmCurve==0){//if led curve
            for (int i=0; i < 8; ++i) 
            {
                m_target_level[i] = (((DMXslice[i]+1)*(DMXslice[i]+1)-1)>>(16-m_pwmRes));//level = (((DMXslice[i]+1)*(DMXslice[i]+1)-1)>>(16-m_pwmRes));
                //ledcWrite(startPin+i, level);
            }
        }
        else{//if straight curve
            for (int i=0; i < 8; ++i) 
            {
                m_target_level[i]= DMXslice[i]<<(m_pwmRes-8);//level= DMXslice[i]<<(m_pwmRes-8);
                //ledcWrite(startPin+i, level);
            }
        } 
    }
}

void Port:: on_off_action(const std::vector<uint8_t> & DMXslice,uint8_t startPin)const
{
    for (int i=0; i < m_nbDmxChannels; ++i) 
    {
        if (DMXslice[i]> m_trigger-1 )
        {
            digitalWrite(Arp[startPin+i], 1 - m_inverted);
        }
        else
        {
            digitalWrite(Arp[startPin+i], m_inverted);   
        }
    } 
}

void Port:: servo_action(const std::vector<uint8_t> & DMXslice,uint8_t startPin)const
{
    uint8_t min=0, max=m_servoAngle;
    if(m_inverted)
    {
        min=m_servoAngle, max=0;
    }
    if(m_servoDetach)
    {
        static uint8_t oldVal[8]={0};
        for (int i=0; i < m_nbDmxChannels; ++i) 
        {           
            if(DMXslice[i] != oldVal[i])
            {
                uint8_t output=i+startPin;
                m_servo[i].attach(Arp[output],output);//(pin,channel)
                m_servo[i].write(map(DMXslice[i], 0, 255, min, max));
            }
            else if(m_servo[i].attached())
            {
                m_servo[i].detach();
            }
            oldVal[i+startPin]=DMXslice[i];   
        }
    }  
    else
    {  
        for (int i=0; i < m_nbDmxChannels; ++i) 
        {
            m_servo[i].write(map(DMXslice[i], 0, 255, min, max));
        }      
    }      
}

void Port:: stepper_action(const std::vector<uint8_t> & DMXslice,uint8_t startPin)const
{
    uint16_t this_stepper_addr=0;
    for (int i=0; i < 4; ++i) 
    {           
        if(m_stp_active[i])
        {

            if(m_stp_dmxRes[i])//16 Bits
            {
                uint16_t position =((DMXslice[this_stepper_addr]<<8) | DMXslice[this_stepper_addr+1]);
                m_stepper[i].refresh(position,DMXslice[this_stepper_addr+2]);
                this_stepper_addr = this_stepper_addr+3;
            }    
            else//8 Bits 
            {
                m_stepper[i].refresh(DMXslice[this_stepper_addr],DMXslice[this_stepper_addr+1]);
                this_stepper_addr = this_stepper_addr+2;
            } 
        }
    }
}


void Port:: CCmotor_action(const std::vector<uint8_t> & DMXslice,uint8_t startPin)const
{
    for(int i=0;i<4;i++)
    { 
        
        uint8_t state = DMXslice[i*2];
        bool direction = (state > 127) ? 0 : 1;
        direction = direction - m_CC_inverted[i];
        uint8_t speed = DMXslice[i*2+1];
        if (speed>0)
        {
            speed=map(speed,1,255,m_cc_threshold[i],255);//map(value, fromLow, fromHigh, toLow, toHigh)

            if (direction) // avant
            {
                ledcWrite(startPin+i*2, speed); 
                ledcWrite(startPin+i*2+1, 0);
             }
            else// arrière
            {
                ledcWrite(startPin+i*2, 0);
                ledcWrite(startPin+i*2+1, speed); 
            }
        }
        else // Stop (freinage)
        {
            ledcWrite(startPin+i*2, 0); 
            ledcWrite(startPin+i*2+1, 0);
        }   
    }
}

void Port:: custom_action(const std::vector<uint8_t> & DMXslice,uint8_t startPin)const
{
    if(m_port_name=='A')
    {
        action_A(DMXslice);
    }
    else if(m_port_name=='B')
    {
        action_B(DMXslice);
    }
}

void Port:: runAtEachLoop()
{
    //if (m_mode==3)//stepper motor
    //{
    switch(m_mode)
    {
        case 0 : {//PWM
            uint8_t startPin= (m_port_name=='A') ? 0 : 8;
            for(uint8_t i=0;i<8;i++)
            {
                int32_t level;
                if(m_actual_level[i] < m_target_level[i])
                {
                    level = m_actual_level[i]+1+(m_actual_level[i]/256);//m_actual_level[i]++;
                    if(level > m_target_level[i])
                            m_actual_level[i] = m_target_level[i];
                    else    m_actual_level[i] = level;
                    ledcWrite(startPin+i, m_actual_level[i]);
                }
                else if(m_actual_level[i] > m_target_level[i])
                {
                    level = m_actual_level[i]-1-(m_actual_level[i]/256);//m_actual_level[i]--;
                    if(level < m_target_level[i])
                            m_actual_level[i] = m_target_level[i];
                    else    m_actual_level[i] = level;                        
                    ledcWrite(startPin+i, m_actual_level[i]);
                }
            }
            break;
        }
        case 3 : {//stepper
            for(uint8_t i=0;i<4;i++)
            {
                if(m_stp_active[i])
                {
                    if(m_action[i]==0)  //perform mode
                    {
                        m_stepper[i].perform();
                    }
                    else                //continuous mode
                    {
                        m_stepper[i].continuous();
                    }                     
                }
            }
            break;
        }
        case 254 : {
            if(m_port_name=='A')
            {
                at_each_loop_A();
            }
            else if(m_port_name=='B')
            {
                at_each_loop_B();
            }
            break;
        }
    }
} 